@@ -9,6 +9,8 @@ from .io import autoopen
 from . import tqdm, HOME_DIR
 import tarfile
 
+import pyterrier
+
 STANDARD_TERRIER_INDEX_FILES = [
     "data.direct.bf",
     "data.document.fsarrayfile",
@@ -21,6 +23,17 @@ STANDARD_TERRIER_INDEX_FILES = [
     "data.properties"
 ]
 
+class GeneratorLen(object):
+    def __init__(self, gen, length):
+        self.gen = gen
+        self.length = length
+
+    def __len__(self): 
+        return self.length
+
+    def __iter__(self):
+        return self.gen
+
 class Dataset():
     """
         Represents a dataset (test collection) for indexing or retrieval. A common use-case is to use the Dataset within an Experiment::
@@ -124,16 +137,17 @@ class RemoteDataset(Dataset):
             raise ValueError("No %s in dataset %s" % (component, name))
         if variant is None:
             if not isinstance(self.locations[component], list):
-                raise ValueError("For %s in dataset %s, you must specify a variant=. Available are: %s" % (component, name, str(self.locations[component].keys())))
+                raise ValueError("For %s in dataset %s, you must specify a variant. Available are: %s" % (component, name, str(list(self.locations[component].keys()))))
             location = self.locations[component][0]
         else:
             if isinstance(self.locations[component], list):
                 raise ValueError("For %s in dataset %s, there are no variants, but you specified %s" % (component, name, variant))
             if not variant in self.locations[component]:
-                raise ValueError("For %s in dataset %s, there is no variant %s. Available are: %s" % (component, name, variant, str(self.locations[component].keys())))
+                raise ValueError("For %s in dataset %s, there is no variant %s. Available are: %s" % (component, name, variant, str(list(self.locations[component].keys()))))
 
     def _get_one_file(self, component, variant=None):
-        filetype=None        
+        filetype=None
+        name=self.name
         self._check_variant(component, variant)
         location = self.locations[component][0] if variant is None else self.locations[component][variant]
 
@@ -150,20 +164,21 @@ class RemoteDataset(Dataset):
         URL = location[1]
         if len(location) > 2:
             filetype = location[2]
-        if "#" in URL:
+
+        if not os.path.exists(self.corpus_home):
+            os.makedirs(self.corpus_home)
+        
+        local = os.path.join(self.corpus_home, local)
+        if "#" in URL and not os.path.exists(local):
             tarname, intarfile = URL.split("#")
             assert not "/" in intarfile
             assert ".tar" in tarname or ".tgz" in tarname
             localtarfile, _ = self._get_one_file("tars", tarname)
             tarobj = tarfile.open(localtarfile, "r")
             tarobj.extract(intarfile, path=self.corpus_home)
-            local = os.path.join(self.corpus_home, local)
             os.rename(os.path.join(self.corpus_home, intarfile), local)
-            return (local, filetype)
-
-        if not os.path.exists(self.corpus_home):
-            os.makedirs(self.corpus_home)
-        local = os.path.join(self.corpus_home, local)
+            return (local, filetype)        
+        
         if not os.path.exists(local):
             try:
                 print("Downloading %s %s to %s" % (self.name, component, local))
@@ -279,11 +294,13 @@ class IRDSDataset(Dataset):
         it = ds.docs_iter()
         if verbose:
             it = tqdm(it, desc=f'{self._irds_id} documents', total=ds.docs_count())
-        for doc in it:
-            doc = doc._asdict()
-            # pyterrier uses "docno"
-            doc['docno'] = doc.pop('doc_id')
-            yield doc
+        def gen():
+            for doc in it:
+                doc = doc._asdict()
+                # pyterrier uses "docno"
+                doc['docno'] = doc.pop('doc_id')
+                yield doc
+        return GeneratorLen(gen(), ds.docs_count())
 
     def get_corpus_lang(self):
         ds = self.irds_ref()
@@ -295,7 +312,7 @@ class IRDSDataset(Dataset):
         # this is only for indices where Terrier provides an index already
         raise NotImplementedError("IRDSDataset doesn't support get_index")
 
-    def get_topics(self, variant=None):
+    def get_topics(self, variant=None, tokenise_query=True):
         """
             Returns the topics, as a dataframe, ready for retrieval. 
         """
@@ -316,6 +333,14 @@ class IRDSDataset(Dataset):
         else:
             print(f'There are multiple query fields available: {qcls._fields[1:]}. To use with pyterrier, provide variant or modify dataframe to add query column.')
 
+        # apply pyterrier tokenisation (otherwise the queries may not play well with batchretrieve)
+        if tokenise_query and 'query' in df:
+            import pyterrier as pt
+            tokeniser = pt.autoclass("org.terrier.indexing.tokenisation.Tokeniser").getTokeniser()
+            def pt_tokenise(text):
+                return ' '.join(tokeniser.getTokens(text))
+            df['query'] = df['query'].apply(pt_tokenise)
+
         return df
 
     def get_topics_lang(self):
@@ -459,7 +484,8 @@ TREC_DEEPLEARNING_DOCS_MSMARCO_FILES = {
         { 
             "train" : ("msmarco-doctrain-qrels.tsv.gz", "https://msmarco.blob.core.windows.net/msmarcoranking/msmarco-doctrain-qrels.tsv.gz"),
             "dev" : ("msmarco-docdev-qrels.tsv.gz", "https://msmarco.blob.core.windows.net/msmarcoranking/msmarco-docdev-qrels.tsv.gz"),
-            "test" : ("2019qrels-docs.txt", "https://trec.nist.gov/data/deep/2019qrels-docs.txt")
+            "test" : ("2019qrels-docs.txt", "https://trec.nist.gov/data/deep/2019qrels-docs.txt"),
+            "test-2020" : ("2020qrels-docs.txt", "https://trec.nist.gov/data/deep/2020qrels-docs.txt")
         },
     "info_url" : "https://microsoft.github.io/msmarco/",
     "corpus_iter" : msmarco_document_generate
@@ -472,18 +498,24 @@ TREC_DEEPLEARNING_PASSAGE_MSMARCO_FILES = {
         { 
             "train" : ("queries.train.tsv", "queries.tar.gz#queries.train.tsv", "singleline"),
             "dev" : ("queries.dev.tsv", "queries.tar.gz#queries.dev.tsv", "singleline"),
+            "dev.small" : ("queries.dev.small.tsv", "collectionandqueries.tar.gz#queries.dev.small.tsv", "singleline"),
             "eval" : ("queries.eval.tsv", "queries.tar.gz#queries.eval.tsv", "singleline"),
+            "eval.small" : ("queries.eval.small.tsv", "collectionandqueries.tar.gz#queries.eval.small.tsv", "singleline"),
             "test-2019" : ("msmarco-test2019-queries.tsv.gz", "https://msmarco.blob.core.windows.net/msmarcoranking/msmarco-test2019-queries.tsv.gz", "singleline"),
+            "test-2020" : ("msmarco-test2020-queries.tsv.gz", "https://msmarco.blob.core.windows.net/msmarcoranking/msmarco-test2020-queries.tsv.gz", "singleline")
         },        
     "tars" : {
         "queries.tar.gz" : ("queries.tar.gz", "https://msmarco.blob.core.windows.net/msmarcoranking/queries.tar.gz"),
-        "collection.tar.gz" : ("collection.tar.gz", "https://msmarco.blob.core.windows.net/msmarcoranking/collection.tar.gz")
+        "collection.tar.gz" : ("collection.tar.gz", "https://msmarco.blob.core.windows.net/msmarcoranking/collection.tar.gz"),
+        "collectionandqueries.tar.gz" : ("collectionandqueries.tar.gz", "https://msmarco.blob.core.windows.net/msmarcoranking/collectionandqueries.tar.gz")
     },
     "qrels" : 
         { 
             "train" : ("qrels.train.tsv", "https://msmarco.blob.core.windows.net/msmarcoranking/qrels.train.tsv"),
             "dev" : ("qrels.dev.tsv", "https://msmarco.blob.core.windows.net/msmarcoranking/qrels.dev.tsv"),
-            "test-2019" : ("2019qrels-docs.txt", "https://trec.nist.gov/data/deep/2019qrels-pass.txt")
+            "test-2019" : ("2019qrels-docs.txt", "https://trec.nist.gov/data/deep/2019qrels-pass.txt"),
+            "test-2020" : ("2020qrels-docs.txt", "https://trec.nist.gov/data/deep/2020qrels-pass.txt"),
+            "dev.small" : ("qrels.dev.small.tsv", "collectionandqueries.tar.gz#qrels.dev.small.tsv"),
         },
     "info_url" : "https://microsoft.github.io/MSMARCO-Passage-Ranking/",
     "corpus_iter" : passage_generate
@@ -762,6 +794,7 @@ VASWANI_FILES = {
     "index":
         [(filename, VASWANI_INDEX_BASE + filename) for filename in STANDARD_TERRIER_INDEX_FILES + ["data.meta-0.fsomapfile"]],
     "info_url" : "http://ir.dcs.gla.ac.uk/resources/test_collections/npl/",
+    "corpus_iter" : lambda dataset, **kwargs : pyterrier.index.treccollection2textgen(dataset.get_corpus(), num_docs=11429, verbose=kwargs.get("verbose", False))
 }
 
 DATASET_MAP = {
