@@ -220,6 +220,26 @@ class Utils:
             run_dict_pytrec_eval[row.qid][row.docno] = int(row.label)
         return(run_dict_pytrec_eval)
 
+    @staticmethod
+    def convert_qrels_to_dataframe(qrels_dict) -> pd.DataFrame:
+        """
+        Convert a qrels dictionary to a dataframe
+
+        Args:
+            qrels_dict(Dict[str, Dict[str, int]]): {qid:{docno:label,},}
+
+        Returns:
+            pd.DataFrame: columns=['qid', 'docno', 'label']
+        """
+        result = {'qid': [], 'docno': [], 'label': []}
+        for qid in qrels_dict:
+            for docno, label in qrels_dict[qid]:
+                result['qid'].append(qid)
+                result['docno'].append(docno)
+                result['label'].append(label)
+
+        return pd.DataFrame(result)
+
     @staticmethod
     def convert_res_to_dict(df):
         """
@@ -254,43 +274,17 @@ class Utils:
             batch_retrieve_results_dict = Utils.convert_res_to_dict(res)
         else:
             batch_retrieve_results_dict = res
+
         if isinstance(qrels, pd.DataFrame):
-            qrels_dic = Utils.convert_qrels_to_dict(qrels)
+            qrels_df = qrels
         else:
-            qrels_dic = qrels
-        req_metrics = set()
-        cutdown = False
-        for m in metrics:
-            if m.startswith("ndcg_cut_"):
-                req_metrics.add("ndcg_cut")
-                cutdown = True
-            elif m.startswith("P_"):
-                req_metrics.add("P")
-                cutdown = True
-            else:
-                req_metrics.add(m)
+            qrels_df = Utils.convert_qrels_to_dataframe(qrels)
+        if len(batch_retrieve_results_dict) == 0:
+            raise ValueError("No results for evaluation")
 
-        evaluator = pytrec_eval.RelevanceEvaluator(qrels_dic, req_metrics)
-
-        result = evaluator.evaluate(batch_retrieve_results_dict)
-        if perquery:
-            if not cutdown:
-                return result
-            # user wanted metrics like ndcg_cut_5, but we had to request ndcg_cut
-            # lets cutout the metrics they didnt want
-
-            # get any arbitrary query
-            q = next(iter(result.keys()))
-            todel=[]
-            for m in result[q]:
-                if not m in metrics:
-                    todel.append(m)
-            for q in result:
-                for m in todel:
-                    del result[q][m]
-            return result
-
-        return Utils.mean_of_measures(result)
+        from .pipelines import _run_and_evaluate
+        _, rtr = _run_and_evaluate(res, None, qrels_df, metrics, perquery=perquery)
+        return rtr
 
     @staticmethod
     def ensure(dictionary, measures, qids):
@@ -302,8 +296,9 @@ class Utils:
         return (dictionary, missing)
 
     @staticmethod
-    def mean_of_measures(result, measures=None):
-        import numpy as np
+    def mean_of_measures(result, measures=None, num_q = None):
+        if len(result) == 0:
+            raise ValueError("No measures received - perhaps qrels and topics had no results in common")
         measures_sum = {}
         mean_dict = {}
         if measures is None:
@@ -317,8 +312,10 @@ class Utils:
             for measure in measures:
                 measure_val = val[measure]
                 measures_sum[measure] = measures_sum.get(measure, 0.0) + measure_val
+        if num_q is None:
+            num_q = len(result.values())
         for measure, value in measures_sum.items():
-            mean_dict[measure] = value / (1 if measure in measures_no_mean else len(result.values()) )
+            mean_dict[measure] = value / (1 if measure in measures_no_mean else num_q)
         return mean_dict
 
     # create a dataframe of string of queries or a list or tuple of strings of queries
